/*
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software. You can use, modify
 * and/ or redistribute the software under the terms of the CeCILL-B license as
 * circulated by CEA, CNRS and INRIA at the following URL "http://www.cecill.info".
 * 
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability.
 * 
 * In this respect, the user's attention is drawn to the risks associated with
 * loading, using, modifying and/or developing or reproducing the software by
 * the user in light of its specific status of free software, that may mean that
 * it is complicated to manipulate, and that also therefore means that it is
 * reserved for developers and experienced professionals having in-depth
 * computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling
 * the security of their systems and/or data to be ensured and, more generally,
 * to use and operate it in the same conditions as regards security.
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.amap.lidar.amapvox.visualization.controllers.loadables.meshes.base;

import com.jogamp.opengl.GL3bc;
import com.jogamp.opengl.GLContext;
import fr.amap.lidar.amapvox.visualization.controllers.loadables.Loadable;
import java.nio.IntBuffer;

/**
 * Used to draw a primitive with some options.
 * @author Tristan Muller (tristan.muller@cirad.fr)
 */
public final class Mesh extends Loadable {
    
    /**
     * The primitive to draw.
     */
    private final Primitive primitive;
    
    /**
     * Which shape to draw the primitive.
     */
    private final DrawingType drawingType;
    
    /**
     * How many primitives to draw.
     */
    private int instances;
    
    /**
     * The size of each pixels in a POINT drawing type.
     */
    private float pointSize;
    
    /**
     * If the back faces are hiden or not (by default true).
     */
    private boolean faceCulling;
    
    /**
     * If the depth mask is used or not (by default true).
     */
    private boolean depthMask;
    
    /**
     * Create a mesh with a primitive and a drawing type.
     * @param primitive Primitive to draw.
     * @param drawingType Which shape to draw the primitive.
     */
    public Mesh(final Primitive primitive, final DrawingType drawingType) {
        this.primitive = primitive;
        this.drawingType = drawingType;
        this.instances = 1;
        this.pointSize = 2f;
        this.faceCulling = true;
        this.depthMask = true;
    }

    /**
     * Specify how many primitives to draw.
     * @param instances 
     */
    public final void setInstances(int instances) {
        this.instances = instances;
    }

    /**
     * Change the size of each pixels in a POINT drawing type.
     * @param pointSize 
     */
    public final void setPointSize(final float pointSize) {
        this.pointSize = pointSize;
    }

    /**
     * If the back faces are hiden or not (by default true).
     * @param faceCulling 
     */
    public final void setFaceCulling(final boolean faceCulling) {
        this.faceCulling = faceCulling;
    }

    /**
     * If the depth mas is used or not (by default true)
     * @param depthMask 
     */
    public final void setDepthMask(final boolean depthMask) {
        this.depthMask = depthMask;
    }

    @Override
    protected final void onLoad() {        
        if (primitive != null) {
            primitive.load();
        }
    }

    @Override
    protected final void onBind() {
        if (primitive != null) {
            primitive.bind();
        }
    }
    
    /**
     * Send the positions of this mesh to the current binded program.
     * @param attribID The shader attribute id to bind to.
     */
    public final void bindPositions(final int attribID) {
        if (primitive != null) {
            primitive.bindPositions(attribID);
        }
    }
    
    /**
     * Send th normals of this mesh to the current binded program.
     * @param attribID The shader attribute id to bind to.
     */
    public final void bindNormals(final int attribID) {
        if (null != drawingType) {
            switch (drawingType) {
                case POINTS:
                    primitive.bindPointsNormals(attribID);
                    break;
                case LINES:
                    primitive.bindLinesNormals(attribID);
                    break;
                case TRIANGLES:
                    primitive.bindTrianglesNormals(attribID);
                    break;
                case QUADS:                        
                    primitive.bindQuadsNormals(attribID);
                    break;
            }
        }
    }
    
    /**
     * Draw this mesh by follow these steps :
     * <ol>
     *     <li>Bind the geometry.</li>
     *     <li>Change the point size.</li>
     *     <li>Update the face culling.</li>
     *     <li>Update the face depth mask.</li>
     *     <li>Draw each geometry instance with the current rendering method.</li>
     *     <li>Unbind the geometry.</li>
     * </ol>
     */
    public final void draw() {
        /* Get opengl 3.0 object */
        final GL3bc gl = GLContext.getCurrentGL().getGL3bc();
        
        if (primitive != null) {
            /* Bind the geometry */
            primitive.bind();
            
            /* Change the size of each pixels in a POINT drawing type */
            gl.glPointSize(pointSize);
            
            /* Hide back faces if face culling is enabled */
            if (faceCulling) {
                gl.glCullFace(GL3bc.GL_BACK);
                gl.glEnable(GL3bc.GL_CULL_FACE);
            } else {
                gl.glDisable(GL3bc.GL_CULL_FACE);
            }
            
            /* Change depth mask */
            gl.glDepthMask(depthMask);
            
            /* Draw the available and selected drawing mode */
            if (null != drawingType) {
                switch (drawingType) {
                    case POINTS:
                        /* Get vertex positions */
                        final IntBuffer pointsBuffer = 
                                primitive.getPointIndexBuffer();
                        /* Draw */
                        gl.glDrawElementsInstanced(GL3bc.GL_POINTS, pointsBuffer.limit(),
                                GL3bc.GL_UNSIGNED_INT, pointsBuffer, instances);
                        break;
                    case LINES:
                        /* Get vertex positions */
                        final IntBuffer linesBuffer =
                                primitive.getLineIndexBuffer();
                        /* Draw */
                        gl.glDrawElementsInstanced(GL3bc.GL_LINES, linesBuffer.limit(),
                               GL3bc.GL_UNSIGNED_INT, linesBuffer, instances);
                        break;
                    case TRIANGLES:
                        /* Get vertex positions */
                        final IntBuffer trianglesBuffer =
                                primitive.getTriangleIndexBuffer();
                        /* Draw */
                        gl.glDrawElementsInstanced(GL3bc.GL_TRIANGLES, trianglesBuffer.limit(),
                                GL3bc.GL_UNSIGNED_INT, trianglesBuffer, instances);
                        break;
                    case QUADS:                        
                        /* Get vertex positions */
                        final IntBuffer quadsBuffer =
                                primitive.getQuadIndexBuffer();
                        /* Draw */
                        gl.glDrawElementsInstanced(GL3bc.GL_QUADS, quadsBuffer.limit(),
                                GL3bc.GL_UNSIGNED_INT, quadsBuffer, instances);
                        break;
                }
            }
            
            /* Unbind the geometry */
            primitive.unbind();
        }
    }

    @Override
    protected void onUnbind() {
        if (primitive != null) {
            primitive.unbind();
        }
    }

    @Override
    protected final void onUnload() {
        if (primitive != null) {
            primitive.unload();
        }
    }
    
}
