/*
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software. You can use, modify
 * and/ or redistribute the software under the terms of the CeCILL-B license as
 * circulated by CEA, CNRS and INRIA at the following URL "http://www.cecill.info".
 * 
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability.
 * 
 * In this respect, the user's attention is drawn to the risks associated with
 * loading, using, modifying and/or developing or reproducing the software by
 * the user in light of its specific status of free software, that may mean that
 * it is complicated to manipulate, and that also therefore means that it is
 * reserved for developers and experienced professionals having in-depth
 * computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling
 * the security of their systems and/or data to be ensured and, more generally,
 * to use and operate it in the same conditions as regards security.
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.amap.lidar.amapvox.visualization.views;

import com.jogamp.nativewindow.WindowClosingProtocol;
import fr.amap.lidar.amapvox.visualization.controllers.voxels.Space;
import com.jogamp.opengl.GL;
import com.jogamp.opengl.GL3bc;
import com.jogamp.opengl.GLAutoDrawable;
import com.jogamp.opengl.GLCapabilities;
import com.jogamp.opengl.GLContext;
import com.jogamp.opengl.GLEventListener;
import com.jogamp.opengl.GLException;
import com.jogamp.opengl.GLProfile;
import com.jogamp.opengl.awt.GLCanvas;
import com.jogamp.opengl.util.Animator;
import fr.amap.lidar.amapvox.visualization.controllers.inputs.ButtonStateType;
import fr.amap.lidar.amapvox.visualization.controllers.inputs.Input;
import fr.amap.lidar.amapvox.visualization.controllers.inputs.Mouse;
import fr.amap.lidar.amapvox.visualization.controllers.voxels.VoxelSpaceFile;
import java.awt.BorderLayout;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import javax.swing.JPanel;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;

/**
 * Class to create the 3D view panel which is used to display each voxels as a
 * group of cubes.
 *
 * @author Tristan Muller (tristan.muller@cirad.fr)
 */
public class Panel3D extends JPanel implements GLEventListener {

    private final GLCanvas glCanvas;
    private Space space;

    private final long startTime;
    private long lastTime;

    private int width;
    private int height;

    public Panel3D() {
        this.glCanvas = initCanvas();
        this.startTime = System.currentTimeMillis();
        this.lastTime = startTime;
        this.width = glCanvas.getSurfaceWidth();
        this.height = glCanvas.getSurfaceHeight();
        this.space = null;
    }

    private GLCanvas initCanvas() {
        try {
            final GLProfile profile = GLProfile.get(GLProfile.GL3bc);
            final GLCapabilities caps = new GLCapabilities(profile);
            caps.setHardwareAccelerated(true);
            caps.setDoubleBuffered(true);
            caps.setSampleBuffers(false);
            
            final GLCanvas canvas = new GLCanvas(caps);
            canvas.addGLEventListener(this);
            canvas.setDefaultCloseOperation(WindowClosingProtocol.WindowClosingMode.DO_NOTHING_ON_CLOSE);

            new Animator(canvas);
            
            canvas.addKeyListener(new KeyAdapter() {
                @Override
                public void keyPressed(KeyEvent e) {
                    Input.setButton(e.getKeyCode(), ButtonStateType.PRESSED);
                }

                @Override
                public void keyReleased(KeyEvent e) {
                    Input.setButton(e.getKeyCode(), ButtonStateType.RELEASED);
                }
            });
            
            canvas.addMouseWheelListener(new MouseWheelListener() {
                @Override
                public void mouseWheelMoved(MouseWheelEvent e) {
                    Input.setScrolled(true);
                    Input.setMouse(new Mouse(e.getX(), e.getY(),
                            -e.getWheelRotation()));
                }
            });
            
            canvas.addMouseMotionListener(new MouseMotionAdapter() {
                int lastX;
                int lastY;
                
                @Override
                public void mouseMoved(MouseEvent e) {
                    Input.setMouse(new Mouse(e.getX() - lastX, e.getY() - lastY,
                            0));

                    lastX = e.getX();
                    lastY = e.getY();
                }

                @Override
                public void mouseDragged(MouseEvent e) {
                    Input.setMouse(new Mouse(e.getX() - lastX, e.getY() - lastY,
                            0));
                    
                    lastX = e.getX();
                    lastY = e.getY();
                }
            });

            canvas.addMouseListener(new MouseAdapter() {
                @Override
                public void mouseReleased(MouseEvent e) {
                    Input.setButton(e.getButton(), ButtonStateType.RELEASED);
                }

                @Override
                public void mousePressed(MouseEvent e) {
                    Input.setButton(e.getButton(), ButtonStateType.PRESSED);
                }
            });

            setLayout(new BorderLayout());
            add(canvas, BorderLayout.CENTER);
            return canvas;
        } catch (final GLException ex) {
            Logger.getLogger(VoxelSpaceFile.class.getName())
                                .log(Level.ERROR, ex, null);
        }

        return null;
    }

    public final GLCanvas getGLCanvas() {
        return glCanvas;
    }

    public final void setSpace(final Space space) {
        this.space = space;
    }

    public final Space getSpace() {
        return space;
    }

    @Override
    public final void init(final GLAutoDrawable drawable) {
        glCanvas.getAnimator().start();
    }

    @Override
    public final void dispose(final GLAutoDrawable drawable) {
        glCanvas.getAnimator().stop();
        space.dispose();
    }

    @Override
    public final void display(final GLAutoDrawable drawable) {
        final GL3bc gl = GLContext.getCurrentGL().getGL3bc();

        final long currentTime = System.currentTimeMillis() - startTime;

        gl.glViewport(0, 0, width, height);

        /* Rescale the scene */
        if (space != null) {
            space.rescale(width, height);
        }

        /* Clear the display */
        gl.glClearColor(0.92f, 0.93f, 0.95f, 1);
        gl.glClear(GL.GL_COLOR_BUFFER_BIT);
        gl.glClear(GL.GL_DEPTH_BUFFER_BIT);
        gl.glEnable(GL.GL_DEPTH_TEST);

        if (space != null) {
            space.display(currentTime - lastTime);
        }

        lastTime = currentTime;
    }

    @Override
    public final void reshape(final GLAutoDrawable drawable, final int x,
            final int y, final int width, final int height) {
        this.width = width;
        this.height = height;
    }

}
